@@ -39,6 +39,17 @@ def cast_tuple(val, depth):
 
 # positional embeddings
 
+class DepthWiseConv1d(nn.Module):
+    def __init__(self, dim_in, dim_out, kernel_size, padding = 0, stride = 1, bias = True, groups = None):
+        super().__init__()
+        groups = default(groups, dim_in)
+        self.net = nn.Sequential(
+            nn.Conv1d(dim_in, dim_in, kernel_size = kernel_size, padding = padding, groups = groups, stride = stride, bias = bias),
+            nn.Conv1d(dim_in, dim_out, 1, bias = bias)
+        )
+    def forward(self, x):
+        return self.net(x)
+
 class FixedPositionalEmbedding(nn.Module):
     def __init__(self, dim):
         super().__init__()
@@ -241,14 +252,24 @@ class Attention(nn.Module):
         dropout = 0.,
         compress_ratio = 1,
         tie_attn_dim = None,
-        rotary_rpe = False
+        rotary_rpe = False,
+        rotary_conv_q_kernel = 15
     ):
         super().__init__()
         inner_dim = dim_head * heads
         self.seq_len = seq_len
         self.heads= heads
         self.scale = dim_head ** -0.5
-        self.to_q = nn.Linear(dim, inner_dim, bias = False)
+
+        if rotary_rpe:
+            self.to_q = nn.Sequential(
+                Rearrange('b n d -> b d n'),
+                DepthWiseConv1d(dim, inner_dim, rotary_conv_q_kernel, padding = (rotary_conv_q_kernel // 2)),
+                Rearrange('b d n -> b n d')
+            )
+        else:
+            self.to_q = nn.Linear(dim, inner_dim, bias = False)
+
         self.to_kv = nn.Linear(dim, inner_dim * 2, bias = False)
         self.to_out = nn.Linear(inner_dim, dim)
 
@@ -607,7 +628,6 @@ class Alphafold2(nn.Module):
         reversible = False,
         sparse_self_attn = False,
         cross_attn_compress_ratio = 1,
-        rotary_rpe = True,
         msa_tie_row_attn = False,
         template_attn_depth = 2,
         num_backbone_atoms = 1,                # number of atoms to reconstitute each residue to, defaults to 3 for C, C-alpha, N
@@ -713,7 +733,7 @@ class Alphafold2(nn.Module):
                 raise ValueError(f'cannot find attention type {attn_type}')
 
             layers.append(nn.ModuleList([
-                prenorm(InterceptAxialAttention(tensor_slice, AxialAttention(dim = dim, template_axial_attn = template_axial_attn, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, sparse_attn = sparse_self_attn, row_attn = row_attn, col_attn = col_attn, rotary_rpe = rotary_rpe))),
+                prenorm(InterceptAxialAttention(tensor_slice, AxialAttention(dim = dim, template_axial_attn = template_axial_attn, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, sparse_attn = sparse_self_attn, row_attn = row_attn, col_attn = col_attn, rotary_rpe = True))),
                 prenorm(InterceptFeedForward(tensor_slice, ff = FeedForward(dim = dim, dropout = ff_dropout))),
                 prenorm(AxialAttention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, tie_row_attn = msa_tie_row_attn, row_attn = row_attn, col_attn = col_attn, rotary_rpe = True)),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
