@@ -238,38 +238,39 @@ def read_msa(filename: str, nseq: int):
 # sidechainnet / MSA / other data utils
 
 def get_esm_embedd(seq, embedd_model, batch_converter, msa_data=None, embedd_type="per_tok"):
-    """ Returns the ESM embeddings for a protein. 
+    """ Returns the ESM or MSA_tr embeddings for a protein. 
         Inputs: 
-        * seq: (L,) tensor of ints (in sidechainnet int-char convention)
-        * embedd_model: ESM model (see train_end2end.py for an example)
-        * batch_converter: ESM batch converter (see train_end2end.py for an example)
+        * seq: ( (b,) L,) tensor of ints (in sidechainnet int-char convention)
+        * embedd_model: ESM/MSA_tr model (see train_end2end.py for an example)
+        * batch_converter: ESM/MSA_tr batch converter (see train_end2end.py for an example)
         * embedd_type: one of ["mean", "per_tok"]. 
-                       "per_tok" is recommended if working with sequences.
+                       "per_tok" is recommended if working with sequences later.
         Outputs: tensor of (batch, n_seqs, L, embedd_dim)
             * n_seqs: number of sequences in the MSA. 1 for ESM-1b
             * embedd_dim: number of embedding dimensions. 
                           768 for MSA_Transformer and 1280 for ESM-1b
     """
-    str_seq = "".join([VOCAB._int2char[x]for x in seq.cpu().numpy()])
+    seq = expand_dims_to(seq, 2 - len(seq.shape))
+    str_seq = ["".join([VOCAB._int2char[x] for x in s]) for s in seq.cpu().numpy()]
     # use MSA transformer
     if msa_data is not None: 
         msa_batch_labels, msa_batch_strs, msa_batch_tokens = batch_converter(msa_data)
         with torch.no_grad():
             results = embedd_model(msa_batch_tokens.to(seq.device), repr_layers=[12], return_contacts=False)
         # index 0 is for start token. so take from 1 one
-        token_reps = results["representations"][12][0, :,  1 : len(str_seq) + 1]
+        token_reps = results["representations"][12][0, :,  1 : len(str_seq[0]) + 1]
         
     # base ESM case
     else: 
-        batch_labels, batch_strs, batch_tokens = batch_converter( [(0, str_seq)] )
+        batch_labels, batch_strs, batch_tokens = batch_converter( [(i, str_seq[i]) for i in range(str_seq[0])] )
         with torch.no_grad():
             results = embedd_model(batch_tokens.to(seq.device), repr_layers=[33], return_contacts=False)
         # index 0 is for start token. so take from 1 one
-        token_reps = results["representations"][33][:, 1 : len(str_seq) + 1].unsqueeze(dim=1)
+        token_reps = results["representations"][33][:, 1 : len(str_seq[0]) + 1].unsqueeze(dim=1)
         
     
     if embedd_type == "mean":
-        token_reps = token_reps.mean(dim=0)
+        token_reps = token_reps.mean(dim=1)
     return token_reps
 
 
@@ -482,13 +483,14 @@ def center_distogram_torch(distogram, bins=DISTANCE_THRESHOLDS, min_t=1., center
     n_bins[-1] = 1.33*bins[-1] # above last threshold is ignored
     max_bin_allowed = torch.tensor(n_bins.shape[0]-1).to(device).long()
     # calculate measures of centrality and dispersion - 
+    magnitudes = distogram.sum(dim=-1)
     if center == "median":
         cum_dist = torch.cumsum(distogram, dim=-1)
-        medium   = 0.5 * torch.ones(*cum_dist.shape[:-1], 1, device=device)
+        medium   = 0.5 * cum_dist[..., -1]
         central  = torch.searchsorted(cum_dist, medium).squeeze()
         central  = n_bins[ torch.min(central, max_bin_allowed) ]
     elif center == "mean":
-        central  = (distogram * n_bins).sum(dim=-1)
+        central  = (distogram * n_bins).sum(dim=-1) / magnitudes
     # create mask for last class - (IGNORE_INDEX)   
     mask = (central <= bins[-2].item()).float()
     # mask diagonal to 0 dist - don't do masked filling to avoid inplace errors
@@ -497,9 +499,9 @@ def center_distogram_torch(distogram, bins=DISTANCE_THRESHOLDS, min_t=1., center
     central[:, diag_idxs, diag_idxs]  *= 0.
     # provide weights
     if wide == "var":
-        dispersion = (distogram * (n_bins - central.unsqueeze(-1))**2).sum(dim=-1)
+        dispersion = (distogram * (n_bins - central.unsqueeze(-1))**2).sum(dim=-1) / magnitudes
     elif wide == "std":
-        dispersion = (distogram * (n_bins - central.unsqueeze(-1))**2).sum(dim=-1).sqrt()
+        dispersion = ((distogram * (n_bins - central.unsqueeze(-1))**2).sum(dim=-1) / magnitudes).sqrt()
     else:
         dispersion = torch.zeros_like(central, device=device)
     # rescale to 0-1. lower std / var  --> weight=1. set potential nan's to 0
@@ -510,7 +512,7 @@ def center_distogram_torch(distogram, bins=DISTANCE_THRESHOLDS, min_t=1., center
 
 # distance matrix to 3d coords: https://github.com/scikit-learn/scikit-learn/blob/42aff4e2e/sklearn/manifold/_mds.py#L279
 
-def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
+def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, eigen=False, verbose=2):
     """ Gets distance matrix. Outputs 3d. See below for wrapper. 
         Assumes (for now) distogram is (N x N) and symmetric
         Outs: 
@@ -519,16 +521,32 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
     """
     device, dtype = pre_dist_mat.device, pre_dist_mat.type()
 
+    # start
+    batch, N, _ = pre_dist_mat.shape
+    diag_idxs = np.arange(N)
+    his = [torch.tensor([np.inf]*batch, device=device)]
+
+    # do it by eigendecomposition - way faster but not weights
+    # https://www.biorxiv.org/content/10.1101/2020.11.27.401232v1.full.pdf
+    if eigen == True and weights is None:
+        preds_3d = []
+        for bi in range(pre_dist_mat.shape[0]):
+            D = pre_dist_mat[bi]**2
+            M = D[:1, :] + D[:, :1] - D 
+            u,s,v = torch.svd_lowrank(M/2)
+            preds_3d.append( (u@torch.diag(s).sqrt())[:, :3].t() )
+        return torch.stack(preds_3d, dim=0), torch.zeros_like(torch.stack(his, dim=0))
+    elif eigen == True:
+        if verbose: 
+            print("Can't use eigen flag if weights are active. Fallback to iterative")
+
+    # continue the iterative way
     if weights is None:
         weights = torch.ones_like(pre_dist_mat)
 
     # ensure batched MDS
     pre_dist_mat = expand_dims_to(pre_dist_mat, length = ( 3 - len(pre_dist_mat.shape) ))
 
-    # start
-    batch, N, _ = pre_dist_mat.shape
-    diag_idxs = np.arange(N)
-    his = [torch.tensor([np.inf]*batch, device=device)]
     # init random coords
     best_stress = float("Inf") * torch.ones(batch, device = device).type(dtype)
     best_3d_coords = 2*torch.rand(batch, N, 3, device = device).type(dtype) - 1
@@ -555,12 +573,13 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
                                                                    stress / dis))
             break
 
+        pre_dist_mat = dist_mat
         best_3d_coords = coords
         his.append( stress / dis )
 
-    return torch.transpose(best_3d_coords, -1,-2), torch.stack(his)
+    return torch.transpose(best_3d_coords, -1,-2), torch.stack(his, dim=0)
 
-def mds_numpy(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
+def mds_numpy(pre_dist_mat, weights=None, iters=10, tol=1e-5, eigen=False, verbose=2):
     """ Gets distance matrix. Outputs 3d. See below for wrapper. 
         Assumes (for now) distrogram is (N x N) and symmetric
         Out:
@@ -867,11 +886,12 @@ def tmscore_numpy(X, Y):
 
 
 def mdscaling_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5,
-                    fix_mirror=True, N_mask=None, CA_mask=None, C_mask=None, verbose=2):
+                    fix_mirror=True, N_mask=None, CA_mask=None, C_mask=None, 
+                    eigen=True, verbose=2):
     """ Handles the specifics of MDS for proteins (mirrors, ...) """
     # batched mds for full parallel 
     preds, stresses = mds_torch(pre_dist_mat, weights=weights,iters=iters, 
-                                              tol=tol, verbose=verbose)
+                                              tol=tol, eigen=True, verbose=verbose)
     if not fix_mirror:
         return preds, stresses
 
